from functools import reduce

from elkm1_lib.message import decode
from elkm1_lib.notify import Notifier


def rx_msg(msg_code: str, msg: str, notifier: Notifier, zeros="00"):
    """
    Create a Elk received message, pass it to decode, and
    invoke the notifiers which will update the base element.
    """
    data = f"{len(msg) + len(zeros) + 4:02X}{msg_code}{msg}{zeros}"
    cksum = (256 - reduce(lambda x, y: x + y, map(ord, data))) % 256
    decoded = decode(f"{data}{cksum:02X}")
    if decoded:
        notifier.notify(decoded[0], decoded[1])



"""Definition of an ElkM1 Task"""

from time import time

from .connection import Connection
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import tn_encode
from .notify import Notifier


class Task(Element):
    """Class representing an Task"""

    def __init__(self, index: int, connection: Connection, notifier: Notifier) -> None:
        super().__init__(index, connection, notifier)
        self.last_change = None

    def activate(self) -> None:
        """(Helper) Activate task"""
        self._connection.send(tn_encode(self._index))


class Tasks(Elements[Task]):
    """Handling for multiple tasks"""

    def __init__(self, connection: Connection, notifier: Notifier) -> None:
        super().__init__(connection, notifier, Task, Max.TASKS.value)
        notifier.attach("TC", self._tc_handler)

    def sync(self) -> None:
        """Retrieve tasks from ElkM1"""
        self.get_descriptions(TextDescriptions.TASK.value)

    def _tc_handler(self, task: int) -> None:
        self.elements[task].setattr("last_change", time(), True)



"""Definition of an ElkM1 Custom Value"""

from __future__ import annotations

from typing import Any

from .connection import Connection
from .const import Max, SettingFormat, TextDescriptions
from .elements import Element, Elements
from .message import cp_encode, cw_encode
from .notify import Notifier


class Setting(Element):
    """Class representing an Custom Value"""

    def __init__(self, index: int, connection: Connection, notifier: Notifier) -> None:
        super().__init__(index, connection, notifier)
        self.value_format = SettingFormat.NUMBER
        self.value = None

    def set(self, value: int | tuple[int, int]) -> None:
        """(Helper) Set custom value."""
        if isinstance(value, tuple):
            if self.value_format != SettingFormat.TIME_OF_DAY:
                raise ValueError("Custom setting 'set' value is wrong format for Elk")
        elif self.value_format == SettingFormat.TIME_OF_DAY:
            raise ValueError("Custom setting 'set' for time of day must have tuple.")
        self._connection.send(cw_encode(self._index, value, self.value_format))


class Settings(Elements[Setting]):
    """Handling for multiple custom values"""

    def __init__(self, connection: Connection, notifier: Notifier) -> None:
        super().__init__(connection, notifier, Setting, Max.SETTINGS.value)
        notifier.attach("CR", self._cr_handler)

    def sync(self) -> None:
        """Retrieve custom values from ElkM1"""
        self._connection.send(cp_encode())
        self.get_descriptions(TextDescriptions.SETTING.value)

    def _cr_handler(self, values: list[dict[str, Any]]) -> None:
        settings = self.elements
        for value in values:
            setting = settings[value["index"]]
            setting.value_format = value["value_format"]
            setting.value = value["value"]



from unittest.mock import Mock

from elkm1_lib.panel import Panel


def test_system_trouble_status_no_trouble():
    panel = Panel(Mock(), Mock())
    panel._ss_handler("0000000000000000000000000000000000")
    assert panel.system_trouble_status == ""


def test_system_trouble_status_boolean_status():
    panel = Panel(Mock(), Mock())
    panel._ss_handler("1000000000000000000000000000000000")
    assert panel.system_trouble_status == "AC Fail"


def test_system_trouble_status_zone_trouble():
    panel = Panel(Mock(), Mock())
    panel._ss_handler("0900000000000000000000000000000000")
    assert panel.system_trouble_status == "Box Tamper zone 9"


def test_system_trouble_status_last_trouble():
    panel = Panel(Mock(), Mock())
    panel._ss_handler("0000000000000000000000000000000005")
    assert panel.system_trouble_status == "Fire zone 5"


def test_system_trouble_status_multiple_troubles():
    panel = Panel(Mock(), Mock())
    panel._ss_handler("1700100000000000000000000000000000")
    assert (
        panel.system_trouble_status == "AC Fail, Box Tamper zone 7, Low Battery Control"
    )



from unittest.mock import Mock

import pytest

from elkm1_lib.const import ThermostatFan, ThermostatMode, ThermostatSetting
from elkm1_lib.message import MessageEncode
from elkm1_lib.thermostats import Thermostat, Thermostats

from .util import rx_msg


@pytest.fixture
def thermostats(notifier):
    return Thermostats(Mock(), notifier)


def test_entry_exit_general(thermostats, notifier):
    rx_msg("TR", "0120072687500", notifier)
    thermostat = thermostats[0]
    assert thermostat.mode == ThermostatMode.COOL
    assert thermostat.hold is False
    assert thermostat.fan == ThermostatFan.AUTO
    assert thermostat.current_temp == 72
    assert thermostat.heat_setpoint == 68
    assert thermostat.cool_setpoint == 75
    assert thermostat.humidity == 0


def test_thermostat_set_types_are_correct(notifier):
    thermostat = Thermostat(0, Mock(), notifier)

    with pytest.raises(ValueError):
        thermostat.set(ThermostatSetting.MODE, 3)
    with pytest.raises(ValueError):
        thermostat.set(ThermostatSetting.COOL_SETPOINT, True)


def test_thermostat_set_sends_correct_command_to_elk():
    mock = Mock()
    thermostat = Thermostat(0, mock, Mock())

    thermostat.set(ThermostatSetting.MODE, ThermostatMode.COOL)
    mock.send.assert_called_with(
        MessageEncode(message="0Bts0102000", response_command=None)
    )

    thermostat.set(ThermostatSetting.HOLD, True)
    mock.send.assert_called_with(
        MessageEncode(message="0Bts0101100", response_command=None)
    )

    thermostat.set(ThermostatSetting.FAN, ThermostatFan.AUTO)
    mock.send.assert_called_with(
        MessageEncode(message="0Bts0100200", response_command=None)
    )

    thermostat.set(ThermostatSetting.HEAT_SETPOINT, 42)
    mock.send.assert_called_with(
        MessageEncode(message="0Bts0142500", response_command=None)
    )



"""Definition of an ElkM1 Light"""

from .connection import Connection
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import pc_encode, pf_encode, pn_encode, ps_encode, pt_encode
from .notify import Notifier


class Light(Element):
    """Class representing a Light"""

    def __init__(self, index: int, connection: Connection, notifier: Notifier) -> None:
        super().__init__(index, connection, notifier)
        self.status = 0

    def level(self, level: int, time: int = 0) -> None:
        """(Helper) Set light to specified level"""
        if level <= 0:
            self._connection.send(pf_encode(self._index))
        elif level >= 98:
            self._connection.send(pn_encode(self._index))
        else:
            self._connection.send(pc_encode(self._index, 9, level, time))

    def toggle(self) -> None:
        """(Helper) Toggle light"""
        self._connection.send(pt_encode(self._index))


class Lights(Elements[Light]):
    """Handling for multiple lights"""

    def __init__(self, connection: Connection, notifier: Notifier) -> None:
        super().__init__(connection, notifier, Light, Max.LIGHTS.value)
        notifier.attach("PC", self._pc_handler)
        notifier.attach("PS", self._ps_handler)

    def sync(self) -> None:
        """Retrieve lights from ElkM1"""
        for i in range(4):
            self._connection.send(ps_encode(i))
        self.get_descriptions(TextDescriptions.LIGHT.value)

    def _pc_handler(self, housecode: str, index: int, light_level: int) -> None:
        self.elements[index].setattr("status", light_level, True)

    def _ps_handler(self, bank: int, statuses: list[int]) -> None:
        for i in range(bank * 64, (bank + 1) * 64):
            self.elements[i].setattr("status", statuses[i - bank * 64], True)



"""
ElkM1 message encode/decode.

Message format:
    LLMM<data>00CC, where:
    LL - length in ASCII hex
    MM - message code
    <data> - message contents; length varies by message code
    00 - constant two characters reserved (almost all messages)
    CC - checksum

The panel numbers zones, keypads, etc. in base 1. The interface of
this module refers to them using base 0. Conversion to base 1 is done
on sending and conversion to base 0 is done on receiving. So, base 0/1
conversion is encapsulated in this module.
"""

from __future__ import annotations

import datetime as dt
import re
import sys
import time
from collections import namedtuple
from collections.abc import Callable
from typing import Any, cast

from .const import (
    AlarmState,
    ArmedStatus,
    ArmLevel,
    ArmUpState,
    ChimeMode,
    ElkRPStatus,
    FunctionKeys,
    Max,
    SettingFormat,
    ThermostatFan,
    ThermostatMode,
    ThermostatSetting,
    ZoneAlarmState,
    ZoneLogicalStatus,
    ZonePhysicalStatus,
    ZoneType,
)

MessageEncode = namedtuple("MessageEncode", ["message", "response_command"])
MsgHandler = Callable[..., None]


def decode(msg: str) -> tuple[str, dict[str, Any]] | None:
    """Decode an Elk message by passing to appropriate decoder"""
    valid, error_msg = _is_valid_length_and_checksum(msg)
    if valid:
        cmd = msg[2:4]
        decoder = getattr(sys.modules[__name__], f"{cmd.lower()}_decode", None)
        if not decoder:
            return ("unknown", {"msg_code": cmd, "data": msg[4:-2]})
        try:
            decoded_msg = decoder(msg)
        except (IndexError, ValueError, AttributeError) as exc:
            raise ValueError("Cannot decode message") from exc
        return (cmd, decoded_msg)

    if not msg or msg.startswith("Username: ") or msg.startswith("Password: "):
        return None
    if "Login successful" in msg:
        return ("login", {"succeeded": True})
    if msg.startswith("Username/Password not found") or msg == "Disabled":
        return ("login", {"succeeded": False})
    raise ValueError(error_msg)


def _is_valid_length_and_checksum(msg: str) -> tuple[bool, str]:
    """Check packet length valid and that checksum is good."""
    try:
        if int(msg[:2], 16) != (len(msg) - 2):
            return (
                False,
                f"Incorrect message length, expected {msg[:2]}, got {len(msg) - 2:02X}. Msg {msg}",  # noqa: E501
            )
        checksum = int(msg[-2:], 16)
        for char in msg[:-2]:
            checksum += ord(char)
        if (checksum % 256) != 0:
            return False, f"Bad checksum. Msg: {msg}"
    except ValueError:
        return False, "Message invalid"

    return True, ""


def _chk_len(msg: str, msg_len: str) -> None:
    if msg[:2] != msg_len:
        raise ValueError(f"Expected msg len {msg_len}. Got msg {msg}")


def am_decode(msg: str) -> dict[str, list[bool]]:
    """AM: Alarm memory by area report."""
    _chk_len(msg, "0C")
    return {"alarm_memory": [x == "1" for x in msg[4 : 4 + Max.AREAS.value]]}


def as_decode(
    msg: str,
) -> dict[str, list[ArmedStatus] | list[ArmUpState] | list[AlarmState]]:
    """AS: Arming status report."""
    return {
        "armed_statuses": [ArmedStatus(x) for x in msg[4:12]],
        "arm_up_states": [ArmUpState(x) for x in msg[12:20]],
        "alarm_states": [AlarmState(x) for x in msg[20:28]],
    }


def az_decode(msg: str) -> dict[str, list[ZoneAlarmState]]:
    """AZ: Alarm by zone report."""
    _chk_len(msg, "D6")
    return {"alarm_status": [ZoneAlarmState(x) for x in msg[4 : 4 + Max.ZONES.value]]}


def _cr_one_custom_value_decode(index: int, part: str) -> dict[str, Any]:
    value = int(part[0:5])
    value_format = SettingFormat(int(part[5]))
    if value_format == SettingFormat.TIME_OF_DAY:
        ret: int | tuple[int, int] = ((value >> 8) & 0xFF, value & 0xFF)
    else:
        ret = value
    return {"index": index, "value": ret, "value_format": value_format}


def cr_decode(msg: str) -> dict[str, Any]:
    """CR: Custom values"""
    if int(msg[4:6]) > 0:
        index = int(msg[4:6]) - 1
        return {"values": [_cr_one_custom_value_decode(index, msg[6:12])]}

    part = 6
    ret = []
    for i in range(Max.SETTINGS.value):
        ret.append(_cr_one_custom_value_decode(i, msg[part : part + 6]))
        part += 6
    return {"values": ret}


def cc_decode(msg: str) -> dict[str, Any]:
    """CC: Output status for single output."""
    return {"output": int(msg[4:7]) - 1, "output_status": msg[7] == "1"}


def cs_decode(msg: str) -> dict[str, Any]:
    """CS: Output status for all outputs."""
    output_status = [x == "1" for x in msg[4 : 4 + Max.OUTPUTS.value]]
    return {"output_status": output_status}


def cv_decode(msg: str) -> dict[str, Any]:
    """CV: Counter value."""
    return {"counter": int(msg[4:6]) - 1, "value": int(msg[6:11])}


def ee_decode(msg: str) -> dict[str, int | bool | ArmedStatus]:
    """EE: Entry/exit timer report."""
    return {
        "area": int(msg[4:5]) - 1,
        "is_exit": msg[5:6] == "0",
        "timer1": int(msg[6:9]),
        "timer2": int(msg[9:12]),
        "armed_status": ArmedStatus(msg[12:13]),
    }


def ic_decode(msg: str) -> dict[str, Any]:
    """IC: Send Valid Or Invalid User Code Format."""
    code = msg[4:16]
    if re.match(r"(0\d){6}", code):
        code = re.sub(r"0(\d)", r"\1", code)
    return {
        "code": code,
        "user": int(msg[16:19]) - 1,
        "keypad": int(msg[19:21]) - 1,
    }


def ie_decode(msg: str) -> dict[str, str]:
    """IE: Installer mode exited."""
    return {}


def ka_decode(msg: str) -> dict[str, Any]:
    """KA: Keypad areas for all keypads."""
    return {"keypad_areas": [ord(x) - 0x31 for x in msg[4 : 4 + Max.KEYPADS.value]]}


def kc_decode(msg: str) -> dict[str, Any]:
    """KC: Keypad key change."""
    return {"keypad": int(msg[4:6]) - 1, "key": int(msg[6:8])}


def kf_decode(msg: str) -> dict[str, Any]:
    """KF: Keypad function key press."""
    return {
        "keypad": int(msg[4:6]) - 1,
        "key": FunctionKeys(msg[6]),
        "chime_mode": [ChimeMode(int(x)) for x in msg[7:15]],
    }


def ld_decode(msg: str) -> dict[str, Any]:
    """LD: System Log Data Update."""
    area = int(msg[11]) - 1
    hour = int(msg[12:14])
    minute = int(msg[14:16])
    month = int(msg[16:18])
    day = int(msg[18:20])
    year = int(msg[24:26]) + 2000
    log_local_datetime = dt.datetime(year, month, day, hour, minute)
    log_local_time = time.mktime(log_local_datetime.timetuple())
    log_gm_timestruct = time.gmtime(log_local_time)

    log: dict[str, Any] = {}
    log["event"] = int(msg[4:8])
    log["number"] = int(msg[8:11])
    log["index"] = int(msg[20:23])
    log["timestamp"] = dt.datetime(*log_gm_timestruct[:6], tzinfo=dt.UTC).isoformat()

    return {"area": area, "log": log}


def lw_decode(msg: str) -> dict[str, Any]:
    """LW: temperatures from all keypads and zones 1-16."""
    keypad_temps = []
    zone_temps = []
    for i in range(16):
        keypad_temps.append(int(msg[4 + 3 * i : 7 + 3 * i]) - 40)
        zone_temps.append(int(msg[52 + 3 * i : 55 + 3 * i]) - 60)
    return {"keypad_temps": keypad_temps, "zone_temps": zone_temps}


def pc_decode(msg: str) -> dict[str, Any]:
    """PC: PLC (lighting) change."""
    housecode = msg[4:7]
    return {
        "housecode": housecode,
        "index": housecode_to_index(housecode),
        "light_level": int(msg[7:9]),
    }


def ps_decode(msg: str) -> dict[str, Any]:
    """PS: PLC (lighting) status."""
    return {
        "bank": ord(msg[4]) - 0x30,
        "statuses": [ord(x) - 0x30 for x in msg[5:69]],
    }


def rp_decode(msg: str) -> dict[str, ElkRPStatus]:
    """RP: Remote programming status."""
    return {"remote_programming_status": ElkRPStatus(int(msg[4:6]))}


def rr_decode(msg: str) -> dict[str, str]:
    """RR: Realtime clock."""
    return {"real_time_clock": msg[4:20]}


def sd_decode(msg: str) -> dict[str, Any]:
    """SD: Description text."""
    desc_ch1 = msg[9]
    show_on_keypad = ord(desc_ch1) >= 0x80
    if show_on_keypad:
        desc_ch1 = chr(ord(desc_ch1) & 0x7F)
    return {
        "desc_type": int(msg[4:6]),
        "unit": int(msg[6:9]) - 1,
        "desc": (desc_ch1 + msg[10:25]).rstrip(),
        "show_on_keypad": show_on_keypad,
    }


def ss_decode(msg: str) -> dict[str, str]:
    """SS: System status."""
    return {"system_trouble_status": msg[4:-2]}


def st_decode(msg: str) -> dict[str, int]:
    """ST: Temperature update."""
    group = int(msg[4:5])
    temperature = int(msg[7:10])
    if group == 0:
        temperature -= 60
    elif group == 1:
        temperature -= 40
    return {"group": group, "device": int(msg[5:7]) - 1, "temperature": temperature}


def tc_decode(msg: str) -> dict[str, int]:
    """TC: Task change."""
    return {"task": int(msg[4:7]) - 1}


def tr_decode(msg: str) -> dict[str, Any]:
    """TR: Thermostat data response."""
    _chk_len(msg, "13")
    return {
        "thermostat_index": int(msg[4:6]) - 1,
        "mode": ThermostatMode(int(msg[6])),
        "hold": msg[7] == "1",
        "fan": ThermostatFan(int(msg[8])),
        "current_temp": int(msg[9:11]),
        "heat_setpoint": int(msg[11:13]),
        "cool_setpoint": int(msg[13:15]),
        "humidity": int(msg[15:17]),
    }


def ua_decode(msg: str) -> dict[str, Any]:
    """UA: Valid User Code Areas."""
    return {
        "user_code": int(msg[4:10]),
        "valid_areas": int(msg[10:12], 16),
        "diagnostic": msg[12:20],
        "user_code_length": int(msg[20]),
        "user_code_type": int(msg[21]),
        "temperature_units": msg[22],
    }


def vn_decode(msg: str) -> dict[str, str]:
    """VN: Version information."""
    elkm1_version = f"{int(msg[4:6], 16)}.{int(msg[6:8], 16)}.{int(msg[8:10], 16)}"
    xep_version = f"{int(msg[10:12], 16)}.{int(msg[12:14], 16)}.{int(msg[14:16], 16)}"
    return {"elkm1_version": elkm1_version, "xep_version": xep_version}


def xk_decode(msg: str) -> dict[str, str]:
    """XK: Ethernet Test."""
    return {"real_time_clock": msg[4:20]}


def zb_decode(msg: str) -> dict[str, Any]:
    """ZB: Zone bypass report."""
    return {"zone_number": int(msg[4:7]) - 1, "zone_bypassed": msg[7] == "1"}


def zc_decode(
    msg: str,
) -> dict[str, int | tuple[ZoneLogicalStatus, ZonePhysicalStatus]]:
    """ZC: Zone Change."""
    _chk_len(msg, "0A")
    status = _status_decode(int(msg[7:8], 16))
    return {"zone_number": int(msg[4:7]) - 1, "zone_status": status}


def zd_decode(msg: str) -> dict[str, list[ZoneType]]:
    """ZD: Zone definitions."""
    _chk_len(msg, "D6")
    zone_definitions = [ZoneType(ord(x) - 0x30) for x in msg[4 : 4 + Max.ZONES.value]]
    return {"zone_definitions": zone_definitions}


def zp_decode(msg: str) -> dict[str, list[int]]:
    """ZP: Zone partitions."""
    zone_partitions = [ord(x) - 0x31 for x in msg[4 : 4 + Max.ZONES.value]]
    return {"zone_partitions": zone_partitions}


def zs_decode(
    msg: str,
) -> dict[str, list[tuple[ZoneLogicalStatus, ZonePhysicalStatus]]]:
    """ZS: Zone statuses."""
    _chk_len(msg, "D6")
    status = [_status_decode(int(x, 16)) for x in msg[4 : 4 + Max.ZONES.value]]
    return {"zone_statuses": status}


def zv_decode(msg: str) -> dict[str, Any]:
    """ZV: Zone voltage."""
    return {"zone_number": int(msg[4:7]) - 1, "zone_voltage": int(msg[7:10]) / 10}


def housecode_to_index(housecode: str) -> int:
    """Convert a X10 housecode to a zero-based index"""
    match = re.search(r"^([A-P])(\d{1,2})$", housecode.upper())
    if match:
        house_index = int(match.group(2))
        if 1 <= house_index <= 16:
            return (ord(match.group(1)) - ord("A")) * 16 + house_index - 1
    raise ValueError(f"Invalid X10 housecode: {housecode}")


def index_to_housecode(index: int) -> str:
    """Convert a zero-based index to a X10 housecode."""
    if index < 0 or index > 255:
        raise ValueError
    quotient, remainder = divmod(index, 16)
    return f"{chr(ord('A') + quotient)}{remainder + 1:02}"


def get_elk_command(line: str) -> str:
    """Return the 2 character command in the message."""
    if len(line) < 4:
        return ""
    return line[2:4]


def _status_decode(status: int) -> tuple[ZoneLogicalStatus, ZonePhysicalStatus]:
    """Decode a 1 byte status into logical and physical statuses."""
    logical_status = ZoneLogicalStatus((status & 0b00001100) >> 2)
    physical_status = ZonePhysicalStatus(status & 0b00000011)
    return (logical_status, physical_status)


def al_encode(arm_mode: ArmLevel, area: int, user_code: int) -> MessageEncode:
    """al: Arm system. Note in 'al' the 'l' can vary"""
    return MessageEncode(f"0Da{arm_mode.value}{area + 1:1}{user_code:06}00", "AS")


def as_encode() -> MessageEncode:
    """as: Get area status."""
    return MessageEncode("06as00", "AS")


def az_encode() -> MessageEncode:
    """az: Get alarm by zone."""
    return MessageEncode("06az00", "AZ")


def cf_encode(output: int) -> MessageEncode:
    """cf: Turn off output."""
    return MessageEncode(f"09cf{output + 1:03}00", None)


def ct_encode(output: int) -> MessageEncode:
    """ct: Toggle output."""
    return MessageEncode(f"09ct{output + 1:03}00", None)


def cn_encode(output: int, seconds: int) -> MessageEncode:
    """cn: Turn on output."""
    return MessageEncode(f"0Ecn{output + 1:03}{seconds:05}00", None)


def cs_encode() -> MessageEncode:
    """cs: Get all output status."""
    return MessageEncode("06cs00", "CS")


def cp_encode() -> MessageEncode:
    """cp: Get ALL custom values."""
    return MessageEncode("06cp00", "CR")


def cr_encode(index: int) -> MessageEncode:
    """cr: Get a custom value."""
    return MessageEncode(f"08cr{index + 1:02}00", "CR")


def cw_encode(
    index: int, value: int | tuple[int, int], value_format: SettingFormat
) -> MessageEncode:
    """cw: Write a custom value."""
    if value_format == SettingFormat.TIME_OF_DAY:
        val = cast(tuple[int, int], value)
        enc = val[0] * 256 + val[1]
    else:
        enc = cast(int, value)
    return MessageEncode(f"0Dcw{index + 1:02}{enc:05}00", None)


def cv_encode(counter: int) -> MessageEncode:
    """cv: Get counter."""
    return MessageEncode(f"08cv{counter + 1:02}00", "CV")


def cx_encode(counter: int, value: int) -> MessageEncode:
    """cx: Change counter value."""
    return MessageEncode(f"0Dcx{counter + 1:02}{value:05}00", "CV")


def dm_encode(
    keypad_area: int, clear: int, beep: bool, timeout: int, line1: str, line2: str
) -> MessageEncode:
    """dm: Display message on keypad."""
    return MessageEncode(
        (
            f"2Edm{keypad_area + 1:1}{clear:1}{beep:1}"
            f"{timeout:05}{line1:^<16.16}{line2:^<16.16}00"
        ),
        None,
    )


def ka_encode() -> MessageEncode:
    """ka: Get keypad areas."""
    return MessageEncode("06ka00", "KA")


def kf_encode(
    keypad: int, functionkey: FunctionKeys = FunctionKeys.FORCE_KF_SYNC
) -> MessageEncode:
    """kf: Function Key Press."""
    return MessageEncode(f"09kf{keypad + 1:02}{functionkey.value}00", "KF")


def lw_encode() -> MessageEncode:
    """lw: Get temperature data."""
    return MessageEncode("06lw00", "LW")


def pc_encode(
    index: int, function_code: int, extended_code: int, seconds: int
) -> MessageEncode:
    """pc: Control any PLC device."""
    return MessageEncode(
        f"11pc{index_to_housecode(index)}{function_code:02}{extended_code:02}{seconds:04}00",
        None,
    )


def pf_encode(index: int) -> MessageEncode:
    """pf: Turn off light."""
    return MessageEncode(f"09pf{index_to_housecode(index)}00", None)


def pn_encode(index: int) -> MessageEncode:
    """pn: Turn on light."""
    return MessageEncode(f"09pn{index_to_housecode(index)}00", None)


def ps_encode(bank: int) -> MessageEncode:
    """ps: Get lighting status."""
    return MessageEncode(f"07ps{bank:1}00", "PS")


def pt_encode(index: int) -> MessageEncode:
    """pt: Toggle light."""
    return MessageEncode(f"09pt{index_to_housecode(index)}00", None)


def sd_encode(desc_type: int, unit: int) -> MessageEncode:
    """sd: Get description."""
    return MessageEncode(f"0Bsd{desc_type:02}{unit + 1:03}00", "SD")


def sp_encode(phrase: int) -> MessageEncode:
    """sp: Speak phrase."""
    return MessageEncode(f"09sp{phrase:03}00", None)


def ss_encode() -> MessageEncode:
    """ss: Get system trouble status."""
    return MessageEncode("06ss00", "SS")


def sw_encode(word: int) -> MessageEncode:
    """sw: Speak word."""
    return MessageEncode(f"09sw{word:03}00", None)


def rw_encode(date_time: dt.datetime) -> MessageEncode:
    """rw: Write time given a datetime."""
    elk_weekday = (date_time.weekday() + 1) % 7 + 1
    time_str = date_time.strftime("%S%M%H.%d%m%y").replace(".", str(elk_weekday))
    return MessageEncode(f"13rw{time_str}00", None)


def tn_encode(task: int) -> MessageEncode:
    """tn: Activate task."""
    return MessageEncode(f"09tn{task + 1:03}00", None)


def tr_encode(thermostat: int) -> MessageEncode:
    """tr: Request thermostat data."""
    return MessageEncode(f"08tr{thermostat + 1:02}00", None)


def ts_encode(thermostat: int, value: int, element: ThermostatSetting) -> MessageEncode:
    """ts: Set thermostat data."""
    return MessageEncode(f"0Bts{thermostat + 1:02}{value:02}{element.value:1}00", None)


def ua_encode(user_code: int) -> MessageEncode:
    """ua: Requst valid user code areas"""
    return MessageEncode(f"0Cua{user_code:06}00", "UA")


def vn_encode() -> MessageEncode:
    """zd: Get panel software version information."""
    return MessageEncode("06vn00", "VN")


def zb_encode(zone: int, area: int, user_code: int) -> MessageEncode:
    """zb: Zone bypass. Zone < 0 unbypass all; Zone > Max bypass all."""
    if zone < 0:
        zone = 0
    elif zone > Max.ZONES.value:
        zone = 999
    else:
        zone += 1
    return MessageEncode(
        f"10zb{zone:03}{area + 1:1}{user_code:06}00",
        "ZB",
    )


def zd_encode() -> MessageEncode:
    """zd: Get zone definitions"""
    return MessageEncode("06zd00", "ZD")


def zp_encode() -> MessageEncode:
    """zp: Get zone partitions"""
    return MessageEncode("06zp00", "ZP")


def zs_encode() -> MessageEncode:
    """zs: Get zone statuses"""
    return MessageEncode("06zs00", "ZS")


def zt_encode(zone: int) -> MessageEncode:
    """zt: Trigger zone."""
    return MessageEncode(f"09zt{zone + 1:03}00", None)


def zv_encode(zone: int) -> MessageEncode:
    """zv: Get zone voltage"""
    return MessageEncode(f"09zv{zone + 1:03}00", "ZV")



"""Utility functions"""

from __future__ import annotations

import ssl
from functools import cache

TLS_VERSIONS = {
    # Unfortunately M1XEP does not support auto-negotiation for TLS
    # protocol; the user code must figure out the version to use. The
    # simplest way is to configure using the connection URL (smarter would
    # be to try to connect using each of the version, except SSL lib does
    # not report TLS error, it just closes the connection, so no easy way to
    # know a different protocol version should be tried)
    "elks": ssl.TLSVersion.TLSv1,
    "elksv1_0": ssl.TLSVersion.TLSv1,
    "elksv1_2": ssl.TLSVersion.TLSv1_2,
    "elksv1_3": ssl.TLSVersion.TLSv1_3,
}


def url_scheme_is_secure(url: str) -> bool:
    """Check if the URL is one that requires SSL/TLS."""
    scheme, _dest = url.split("://")
    return scheme.startswith("elks")


@cache
def ssl_context_for_scheme(scheme: str) -> ssl.SSLContext:
    """Create an SSL context for the given scheme.

    Since ssl context is expensive to create, cache it
    for future use since we only have a few schemes.
    """
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    if tls := TLS_VERSIONS.get(scheme):
        ssl_context.minimum_version = tls
        ssl_context.maximum_version = tls

    ssl_context.check_hostname = False
    ssl_context.verify_mode = ssl.CERT_NONE
    ssl_context.set_ciphers("DEFAULT:@SECLEVEL=0")

    # ssl.OP_LEGACY_SERVER_CONNECT is only available in Python 3.12a4+
    ssl_context.options |= getattr(ssl, "OP_LEGACY_SERVER_CONNECT", 0x4)
    return ssl_context


def parse_url(url: str) -> tuple[str, str, int, ssl.SSLContext | None]:
    """Parse a Elk connection string"""
    scheme, dest = url.split("://")
    host = None
    ssl_context = None
    if scheme == "elk":
        host, port = dest.split(":") if ":" in dest else (dest, "2101")
    elif TLS_VERSIONS.get(scheme):
        host, port = dest.split(":") if ":" in dest else (dest, "2601")
        ssl_context = ssl_context_for_scheme(scheme)
        scheme = "elks"
    elif scheme == "serial":
        host, port = dest.split(":") if ":" in dest else (dest, "115200")
    else:
        raise ValueError(f"Invalid scheme '{scheme}'")
    return (scheme, host, int(port), ssl_context)


def pretty_const(value: str) -> str:
    """Make a constant pretty for printing in GUI"""
    words = value.split("_")
    pretty = words[0].capitalize()
    for word in words[1:]:
        pretty += " " + word.lower()
    return pretty



"""
Base of all the elements found on the Elk panel... Zone, Keypad, etc.
"""

from __future__ import annotations

import re
from abc import abstractmethod
from collections.abc import Callable, Generator
from typing import Any, Generic, TypeVar

from .connection import Connection
from .const import TextDescription, TextDescriptions
from .message import sd_encode
from .notify import Notifier


class Element:
    """Element class"""

    def __init__(self, index: int, connection: Connection, notifier: Notifier) -> None:
        self._index = index
        self._connection = connection
        self._notifier = notifier
        self._observers: list[Callable[[Element, dict[str, Any]], None]] = []
        self.name: str = self.default_name()
        self._changeset: dict[str, Any] = {}
        self._configured: bool = False

    @property
    def index(self) -> int:
        """Get the index, immutable once class created"""
        return self._index

    @property
    def configured(self) -> bool:
        """If a callback has ever been triggered this will be true."""
        return self._configured

    def add_callback(self, observer: Callable[[Element, dict[str, Any]], None]) -> None:
        """Callbacks when attribute of element changes"""
        self._observers.append(observer)

    def remove_callback(
        self, observer: Callable[[Element, dict[str, Any]], None]
    ) -> None:
        """Callbacks when attribute of element changes"""
        if observer in self._observers:
            self._observers.remove(observer)

    def _notify(self) -> None:
        """Callbacks when attribute of element changes"""
        for observer in self._observers:
            observer(self, self._changeset)
        self._changeset = {}

    def setattr(
        self, attr: str, new_value: Any, close_the_changeset: bool = True
    ) -> None:
        """If attribute value has changed then set it and call the callbacks"""
        existing_value: Any = getattr(self, attr, None)
        if existing_value != new_value:
            setattr(self, attr, new_value)
            self._changeset[attr] = new_value

        if close_the_changeset and self._changeset:
            self._notify()

    def default_name(self, separator: str = "-") -> str:
        """Return a default name for based on class and index of element"""
        return f"{self.__class__.__name__}{separator}{self._index + 1:03}"

    def is_default_name(self) -> bool:
        """Check if the name assigned is the default_name"""
        return self.name == self.default_name()

    def __str__(self) -> str:
        varlist = {
            k: v
            for (k, v) in vars(self).items()
            if not k.startswith("_") and k != "name"
        }.items()
        varstr = " ".join(
            # pylint: disable=consider-using-f-string
            "%s:%s" % item  # noqa
            for item in varlist
        )
        return f"{self._index} '{self.name}' {varstr}"

    def as_dict(self) -> dict[str, Any]:
        """Package up the public attributes as a dict."""
        attrs = vars(self)
        return {key: attrs[key] for key in attrs if not key.startswith("_")}

    def _configured_was_set(self) -> None:
        """Called when configured flag is set for an element."""


T = TypeVar("T", bound=Element)


class Elements(Generic[T]):
    """Base for list of elements."""

    def __init__(
        self,
        connection: Connection,
        notifier: Notifier,
        class_: type[T],
        max_elements: int,
    ) -> None:
        self._connection = connection
        self._notifier = notifier
        self.max_elements = max_elements
        self.elements = [class_(i, connection, notifier) for i in range(max_elements)]

        self._text_desc: TextDescription | None = None
        notifier.attach("SD", self._sd_handler)

    def __iter__(self) -> Generator[Element, None, None]:
        yield from self.elements

    def __getitem__(self, key: int) -> Element:
        return self.elements[key]

    def get_descriptions(self, text_desc: TextDescription) -> None:
        """Gets the descriptions for specified type."""
        self._text_desc = text_desc
        self._connection.send(sd_encode(text_desc.desc_type, 0))

    def _sd_handler(
        self, desc_type: int, unit: int, desc: str, show_on_keypad: bool
    ) -> None:
        if not self._text_desc or desc_type != self._text_desc.desc_type:
            return
        if unit < 0 or unit >= self._text_desc.number_descriptions:
            self._text_desc = None
            return

        if desc_type != TextDescriptions.USER.value.desc_type or not re.match(
            r"USER \d\d\d$", desc
        ):
            element = self.elements[unit]
            element.setattr("name", desc, True)
            element._configured = True  # pylint: disable=protected-access
            element._configured_was_set()  # pylint: disable=protected-access
        self._connection.send(sd_encode(desc_type, unit + 1), priority_send=True)

    @abstractmethod
    def sync(self) -> None:
        """Synchronize elements"""



"""Definition of an ElkM1 Custom Value"""

from __future__ import annotations

from .connection import Connection
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import cv_encode, cx_encode
from .notify import Notifier


class Counter(Element):
    """Class representing an Counter"""

    def __init__(self, index: int, connection: Connection, notifier: Notifier) -> None:
        super().__init__(index, connection, notifier)
        self.value = None

    def get(self) -> None:
        """(Helper) Get counter"""
        self._connection.send(cv_encode(self._index))

    def set(self, value: int) -> None:
        """(Helper) Set counter to value"""
        self._connection.send(cx_encode(self._index, value))

    def _configured_was_set(self) -> None:
        self._connection.send(cv_encode(self.index), priority_send=True)


class Counters(Elements[Counter]):
    """Handling for multiple counters"""

    def __init__(self, connection: Connection, notifier: Notifier) -> None:
        super().__init__(connection, notifier, Counter, Max.COUNTERS.value)
        notifier.attach("CV", self._cv_handler)

    def sync(self) -> None:
        """Retrieve values from ElkM1 on demand"""
        self.get_descriptions(TextDescriptions.COUNTER.value)

    def _cv_handler(self, counter: int, value: int) -> None:
        self.elements[counter].setattr("value", value, True)



"""Definition of an ElkM1 Thermostat"""

from __future__ import annotations

from .connection import Connection
from .const import (
    Max,
    TextDescriptions,
    ThermostatFan,
    ThermostatMode,
    ThermostatSetting,
)
from .elements import Element, Elements
from .message import tr_encode, ts_encode
from .notify import Notifier

SETTING_TYPING = {
    ThermostatSetting.MODE: ThermostatMode,
    ThermostatSetting.HOLD: bool,
    ThermostatSetting.FAN: ThermostatFan,
    ThermostatSetting.GET_TEMPERATURE: int,
    ThermostatSetting.COOL_SETPOINT: int,
    ThermostatSetting.HEAT_SETPOINT: int,
}


class Thermostat(Element):
    """Class representing an Thermostat"""

    def __init__(self, index: int, connection: Connection, notifier: Notifier) -> None:
        super().__init__(index, connection, notifier)
        self.mode: ThermostatMode | None = None
        self.hold = False
        self.fan: ThermostatFan | None = None
        self.current_temp = 0
        self.heat_setpoint = 0
        self.cool_setpoint = 0
        self.humidity = 0

    def set(
        self,
        element_to_set: ThermostatSetting,
        val: bool | int | ThermostatMode | ThermostatFan,
    ) -> None:
        """(Helper) Set thermostat"""
        if (  # pylint: disable=unidiomatic-typecheck
            type(val) is not SETTING_TYPING[element_to_set]
        ):
            raise ValueError("Wrong type for thermostat setting.")
        if isinstance(val, bool):
            setting = 1 if val else 0
        elif isinstance(val, ThermostatFan | ThermostatMode):
            setting = val.value
        else:
            setting = val

        self._connection.send(ts_encode(self.index, setting, element_to_set))

    def _configured_was_set(self) -> None:
        self._connection.send(tr_encode(self.index), priority_send=True)


class Thermostats(Elements[Thermostat]):
    """Handling for multiple areas"""

    def __init__(self, connection: Connection, notifier: Notifier) -> None:
        super().__init__(connection, notifier, Thermostat, Max.THERMOSTATS.value)
        notifier.attach("ST", self._st_handler)
        notifier.attach("TR", self._tr_handler)

    def sync(self) -> None:
        """Retrieve areas from ElkM1"""
        self.get_descriptions(TextDescriptions.THERMOSTAT.value)

    def _st_handler(self, group: int, device: int, temperature: int) -> None:
        if group == 2:
            self.elements[device].setattr("current_temp", temperature, True)

    def _tr_handler(
        self,
        thermostat_index: int,
        mode: ThermostatMode,
        hold: bool,
        fan: ThermostatFan,
        current_temp: int,
        heat_setpoint: int,
        cool_setpoint: int,
        humidity: int,
    ) -> None:
        thermostat = self.elements[thermostat_index]
        thermostat.setattr("mode", mode, False)
        thermostat.setattr("hold", hold, False)
        thermostat.setattr("fan", fan, False)
        thermostat.setattr("current_temp", current_temp, False)
        thermostat.setattr("heat_setpoint", heat_setpoint, False)
        thermostat.setattr("cool_setpoint", cool_setpoint, False)
        thermostat.setattr("humidity", humidity, True)



from unittest.mock import Mock

from elkm1_lib.notify import Notifier


def test_attach():
    mock_notified = Mock()
    notifier = Notifier()
    notifier.attach("foo", mock_notified)
    notifier.notify("foo", {"something": 42})
    mock_notified.assert_called_once_with(something=42)


def test_attach_multiple():
    mock_notified1 = Mock()
    mock_notified2 = Mock()
    notifier = Notifier()
    notifier.attach("foo", mock_notified1)
    notifier.attach("foo", mock_notified2)
    notifier.notify("foo", {"something": 42})
    mock_notified1.assert_called_once_with(something=42)
    mock_notified2.assert_called_once_with(something=42)


def test_attach_same_twice_doesnt_call_twice():
    mock_notified = Mock()
    notifier = Notifier()
    notifier.attach("foo", mock_notified)
    notifier.attach("foo", mock_notified)
    notifier.notify("foo", {"something": 42})
    mock_notified.assert_called_once_with(something=42)


def test_deattach():
    mock_notified1 = Mock()
    mock_notified2 = Mock()
    notifier = Notifier()
    notifier.attach("foo", mock_notified1)
    notifier.attach("foo", mock_notified2)
    notifier.detach("foo", mock_notified1)
    notifier.notify("foo", {"something": 42})
    mock_notified1.assert_not_called()
    mock_notified2.assert_called_once_with(something=42)



import asyncio
import contextlib
from unittest.mock import MagicMock, patch

import pytest

from elkm1_lib.discovery import AIOELKDiscovery, ELKDiscovery, ElkSystem


@pytest.fixture
async def mock_discovery_aio_protocol():
    """Fixture to mock an asyncio connection."""
    loop = asyncio.get_running_loop()
    future = asyncio.Future()

    async def _wait_for_connection():
        transport, protocol = await future
        await asyncio.sleep(0)
        await asyncio.sleep(0)
        return transport, protocol

    async def _mock_create_datagram_endpoint(func, sock=None):
        protocol: ELKDiscovery = func()
        transport = MagicMock()
        protocol.connection_made(transport)
        with contextlib.suppress(asyncio.InvalidStateError):
            future.set_result((transport, protocol))
        return transport, protocol

    with patch.object(loop, "create_datagram_endpoint", _mock_create_datagram_endpoint):
        yield _wait_for_connection


@pytest.mark.asyncio
async def test_async_scanner_specific_address(mock_discovery_aio_protocol):
    """Test scanner with a specific address."""
    scanner = AIOELKDiscovery()

    task = asyncio.ensure_future(
        scanner.async_scan(timeout=10, address="192.168.209.56")
    )
    _, protocol = await mock_discovery_aio_protocol()
    protocol.datagram_received(
        b"M1XEP\x00@\x9d\xb1\xad\x03\xc0\xa8\xd18\n)\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01",
        ("192.168.209.56", 2362),
    )
    await task
    assert scanner.found_devices == [
        ElkSystem(
            mac_address="00:40:9D:B1:AD:03", ip_address="192.168.209.56", port=2601
        )
    ]


@pytest.mark.asyncio
async def test_async_scanner_broadcast(mock_discovery_aio_protocol):
    """Test scanner with a broadcast."""
    scanner = AIOELKDiscovery()

    task = asyncio.ensure_future(scanner.async_scan(timeout=0.01))
    _, protocol = await mock_discovery_aio_protocol()
    protocol.datagram_received(
        b"M1XEP\x00@\x9d\xb1\xad\x03\xc0\xa8\xd18\n)\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01",
        ("192.168.209.56", 2362),
    )
    await task
    assert scanner.found_devices == [
        ElkSystem(
            mac_address="00:40:9D:B1:AD:03", ip_address="192.168.209.56", port=2601
        )
    ]



"""Manage connection and IO to Elk."""

from __future__ import annotations

import asyncio
import logging
from asyncio import timeout as asyncio_timeout
from collections import deque
from functools import reduce
from typing import Any, NamedTuple

from serial_asyncio_fast import open_serial_connection

from .message import MessageEncode, decode, get_elk_command
from .notify import Notifier
from .util import parse_url

LOG = logging.getLogger(__name__)
HEARTBEAT_TIME = 120
MESSAGE_RESPONSE_TIME = 5.0


class QueuedWrite(NamedTuple):
    """Structure for entries in the write queue."""

    msg: str
    response_cmd: str | None
    timeout: float = 5.0
    raw: bool = False


class Connection:
    """Manage connection to ElkM1 panel."""

    def __init__(self, url: str, notifier: Notifier):
        self._url = url
        self._notifier = notifier

        self._writer: asyncio.StreamWriter | None = None
        self._awaiting_response_command: str | None = None
        self._paused = False
        self._write_queue: deque[QueuedWrite] = deque()
        self._check_write_queue = asyncio.Event()
        self._response_received = asyncio.Event()
        self._heartbeat_event = asyncio.Event()
        self._tasks: set[asyncio.Task[Any]] = set()

    async def connect(self) -> None:
        """Create connection to Elk."""

        LOG.info("Connecting to ElkM1 at %s", self._url)
        retry_time = 1
        scheme, dest, param, ssl_context = parse_url(self._url)
        while not self._writer:
            try:
                async with asyncio_timeout(30):
                    if scheme == "serial":
                        reader, self._writer = await open_serial_connection(
                            url=dest, baudrate=param
                        )
                    else:
                        reader, self._writer = await asyncio.open_connection(
                            host=dest, port=param, ssl=ssl_context
                        )
            except (TimeoutError, ValueError, OSError) as err:
                LOG.warning(
                    "Error connecting to ElkM1 (%s). Retrying in %d seconds",
                    err,
                    retry_time,
                )
                await asyncio.sleep(retry_time)
                retry_time = min(60, retry_time * 2)
                continue

            if scheme != "serial":
                self._tasks.add(asyncio.create_task(self._heartbeat_timer()))
            self._tasks.add(asyncio.create_task(self._read_stream(reader)))
            self._tasks.add(asyncio.create_task(self._write_stream()))
            self._notifier.notify("connected", {})

    async def _read_stream(self, reader: asyncio.StreamReader) -> None:
        read_buffer = ""
        while True:
            data = await reader.read(500)
            if not data:
                break
            self._heartbeat()

            read_buffer += data.decode("ISO-8859-1")
            while "\r\n" in read_buffer:
                line, read_buffer = read_buffer.split("\r\n", 1)
                if get_elk_command(line) == self._awaiting_response_command:
                    self._response_received.set()

                LOG.debug("got_data '%s'", line)
                try:
                    decoded = decode(line)
                    if decoded:
                        self._notifier.notify(decoded[0], decoded[1])
                except (ValueError, AttributeError) as exc:
                    LOG.error("Invalid message '%s'", data, exc_info=exc)

    async def _write_stream(self) -> None:
        async def write_msg() -> None:
            if not q_entry.raw:  # pylint: disable=possibly-used-before-assignment
                cksum = (256 - reduce(lambda x, y: x + y, map(ord, q_entry.msg))) % 256
                msg = f"{q_entry.msg}{cksum:02X}\r\n"
            else:
                msg = q_entry.msg + "\r\n"
            LOG.debug("write_data '%s'", msg[:-2])
            self._writer.write((msg).encode())  # type: ignore

        async def await_msg_response() -> None:
            self._awaiting_response_command = q_entry.response_cmd
            try:
                async with asyncio_timeout(MESSAGE_RESPONSE_TIME):
                    await self._response_received.wait()
            except TimeoutError:
                self._notifier.notify("timeout", {"msg_code": q_entry.response_cmd})
            self._response_received.clear()
            self._awaiting_response_command = None

        while True:
            if not self._write_queue:
                await self._check_write_queue.wait()
            if not self._writer:
                break
            self._check_write_queue.clear()
            if self._write_queue:
                q_entry = self._write_queue.popleft()
                await write_msg()
                if q_entry.response_cmd:
                    await await_msg_response()

    def _send(self, q_entry: QueuedWrite, priority_send: bool) -> None:
        if self._paused:
            return
        if priority_send:
            self._write_queue.appendleft(q_entry)
        else:
            self._write_queue.append(q_entry)
        self._check_write_queue.set()

    def send(self, msg: MessageEncode, priority_send: bool = False) -> None:
        """Send a message to Elk."""
        self._send(QueuedWrite(msg.message, msg.response_command), priority_send)

    def send_raw(self, msg: str) -> None:
        """Send a raw message to Elk (no checksum will be added)."""
        self._send(QueuedWrite(msg, None, raw=True), False)

    def is_connected(self) -> bool:
        """Is the connection active?"""
        return self._writer is not None

    def pause(self) -> None:
        """Pause the connection from sending/receiving."""
        self._write_queue.clear()
        self._paused = True

    def resume(self) -> None:
        """Restart the connection sending/receiving."""
        self._paused = False

    def is_paused(self) -> bool:
        """Is the connection paused."""
        return self._paused

    def disconnect(self, reason: str = "") -> None:
        """Disconnect and cleanup."""
        LOG.warning("ElkM1 at %s disconnecting %s", self._url, reason)
        if self._writer:
            self._writer.close()
            self._writer = None
        for task in self._tasks:
            if asyncio.current_task() != task:
                task.cancel()
        self._tasks = set()
        self._notifier.notify("disconnected", {})

    def _heartbeat(self) -> None:
        self._heartbeat_event.set()

    async def _heartbeat_timer(self) -> None:
        while True:
            self._heartbeat_event.clear()
            try:
                async with asyncio_timeout(HEARTBEAT_TIME):
                    await self._heartbeat_event.wait()
            except TimeoutError:
                if self._paused:
                    continue
                self.disconnect("(heartbeat timeout)")
                await self.connect()
                break



import asyncio
import logging
import pprint

from elkm1_lib.discovery import AIOELKDiscovery

logging.basicConfig(level=logging.DEBUG)


async def go():
    scanner = AIOELKDiscovery()
    pprint.pprint(await scanner.async_scan())


asyncio.run(go())



import datetime as dt

import pytest

import elkm1_lib.message as m
from elkm1_lib.const import ArmLevel, SettingFormat, ThermostatSetting


def test_housecode_to_index_accepts_valid_codes():
    assert m.housecode_to_index("A01") == 0
    assert m.housecode_to_index("P16") == 255
    assert m.housecode_to_index("f6") == 85


def test_housecode_to_index_raises_error_on_invalid():
    with pytest.raises(ValueError):
        m.housecode_to_index("asdf")
    with pytest.raises(ValueError):
        m.housecode_to_index("Q01")
    with pytest.raises(ValueError):
        m.housecode_to_index("A00")
    with pytest.raises(ValueError):
        m.housecode_to_index("A17")


def test_index_to_housecode_accepts_valid_indices():
    assert m.index_to_housecode(0) == "A01"
    assert m.index_to_housecode(10) == "A11"
    assert m.index_to_housecode(16) == "B01"
    assert m.index_to_housecode(255) == "P16"


def test_index_to_housecode_raises_error_on_invalid():
    with pytest.raises(ValueError):
        m.index_to_housecode(-1)
    with pytest.raises(ValueError):
        m.index_to_housecode(256)


def test_decode_raises_value_error_on_bad_message():
    with pytest.raises(ValueError):
        m.decode("a really really bad message")


def test_decode_calls_unknown_handler_on_bad_command_or_not_implemented():
    decoded = m.decode("08XXtest28")
    assert decoded
    assert decoded[0] == "unknown"
    assert decoded[1] == {"msg_code": "XX", "data": "test"}


def test_decode_raises_value_error_on_length_too_long():
    with pytest.raises(ValueError) as excinfo:
        m.decode("42CV01000990030")
    assert str(excinfo.value).startswith("Incorrect message length")


def test_decode_raises_value_error_on_length_too_short():
    with pytest.raises(ValueError) as excinfo:
        m.decode("02CV01000990030")
    assert str(excinfo.value).startswith("Incorrect message length")


def test_decode_raises_value_error_on_bad_checksum():
    with pytest.raises(ValueError) as excinfo:
        m.decode("0DCV01000990042")
    assert str(excinfo.value).startswith("Bad checksum")


def test_decode_raises_value_error_on_short_message():
    with pytest.raises(ValueError) as excinfo:
        m.decode("X")
    assert str(excinfo.value) == "Message invalid"


def test_decode_login_success():
    decoded = m.decode("Login successful")
    assert decoded
    assert decoded[0] == "login"
    assert decoded[1] == {"succeeded": True}


def test_decode_login_failed():
    decoded = m.decode("Username/Password not found")
    assert decoded
    assert decoded[0] == "login"
    assert decoded[1] == {"succeeded": False}


def test_encode_message_with_a_variable():
    assert m.ps_encode(1) == ("07ps100", "PS")


def test_al_encode():
    assert m.al_encode(ArmLevel.ARMED_NIGHT, 2, 4242) == ("0Da4300424200", "AS")


def test_cf_encode():
    assert m.cf_encode(42) == ("09cf04300", None)


def test_ct_encode():
    assert m.ct_encode(42) == ("09ct04300", None)


def test_cn_encode():
    assert m.cn_encode(42, 4242) == ("0Ecn0430424200", None)


def test_cr_encode():
    assert m.cr_encode(42) == ("08cr4300", "CR")


def test_cw_encode():
    assert m.cw_encode(5, (21, 40), SettingFormat.TIME_OF_DAY) == (
        "0Dcw060541600",
        None,
    )


def test_cv_encode():
    assert m.cv_encode(12) == ("08cv1300", "CV")


def test_cx_encode():
    assert m.cx_encode(10, 54321) == ("0Dcx115432100", "CV")


def test_dm_encode():
    assert m.dm_encode(4, 1, False, 42, "Hello", "  World") == (
        "2Edm51000042Hello^^^^^^^^^^^  World^^^^^^^^^00",
        None,
    )


def test_dm_encode_too_long_is_truncated():
    assert m.dm_encode(4, 1, False, 42, "Hello this is a long message", "  World") == (
        "2Edm51000042Hello this is a   World^^^^^^^^^00",
        None,
    )


def test_pc_encode():
    assert m.pc_encode(10, 29, 38, 1234) == ("11pcA112938123400", None)


def test_pf_encode():
    assert m.pf_encode(10) == ("09pfA1100", None)


def test_pn_encode():
    assert m.pn_encode(10) == ("09pnA1100", None)


def test_ps_encode():
    assert m.ps_encode(2) == ("07ps200", "PS")


def test_pt_encode():
    assert m.pt_encode(10) == ("09ptA1100", None)


def test_sd_encode():
    assert m.sd_encode(10, 42) == ("0Bsd1004300", "SD")


def test_sp_encode():
    assert m.sp_encode(242) == ("09sp24200", None)


def test_sw_encode():
    assert m.sw_encode(424) == ("09sw42400", None)


def test_rw_encode():
    assert m.rw_encode(dt.datetime(2020, 10, 19, 11, 47, 27, 942490)) == (
        "13rw274711219102000",
        None,
    )


def test_tn_encode():
    assert m.tn_encode(142) == ("09tn14300", None)


def test_tr_encode():
    assert m.tr_encode(6) == ("08tr0700", None)


def test_ts_encode():
    assert m.ts_encode(3, 7, ThermostatSetting.MODE) == ("0Bts0407000", None)


def test_zb_encode():
    assert m.zb_encode(142, 7, 123456) == ("10zb143812345600", "ZB")


def test_zt_encode():
    assert m.zt_encode(199) == ("09zt20000", None)


def test_zv_encode():
    assert m.zv_encode(200) == ("09zv20100", "ZV")


def test_ua_encode():
    assert m.ua_encode(654321) == ("0Cua65432100", "UA")



from unittest.mock import Mock

import pytest

from elkm1_lib.const import ZoneLogicalStatus, ZonePhysicalStatus, ZoneType
from elkm1_lib.zones import Zones

from .util import rx_msg


@pytest.fixture
def zones(notifier):
    return Zones(Mock(), notifier)


def test_zone_change(zones, notifier):
    rx_msg("ZC", "001B", notifier)
    zone = zones[0]
    assert zone.logical_status == ZoneLogicalStatus.VIOLATED
    assert zone.physical_status == ZonePhysicalStatus.SHORT


def test_bad_zone_change_length(notifier):
    with pytest.raises(ValueError):
        rx_msg("ZC", "001", notifier)


def test_zone_definition(zones, notifier):
    rx_msg("ZD", f"1111@00Q{'0' * 200}", notifier)
    assert zones[0].definition == ZoneType.BURGLAR_ENTRY_EXIT_1
    assert zones[4].definition == ZoneType.NON_ALARM
    assert zones[5].definition == ZoneType.DISABLED
    assert zones[7].definition == ZoneType.TEMPERATURE


def test_zone_alarm_state(zones, notifier):
    rx_msg("AZ", f"05:A0000{'0' * 200}", notifier)
    assert zones[0].triggered_alarm is False
    assert zones[1].triggered_alarm is True
    assert zones[2].triggered_alarm is True
    assert zones[3].triggered_alarm is True


def test_zone_voltage(zones, notifier):
    rx_msg("ZV", "123072", notifier)
    assert zones[122].voltage == pytest.approx(7.2)



"""Observer for notifying when messages received or events occur."""

import logging
from collections.abc import Callable
from typing import Any

NotifyHandler = Callable[..., None]
LOG = logging.getLogger(__name__)


class Notifier:
    """Register and notify on events."""

    def __init__(self) -> None:
        """Initialize a new notify instance."""
        self._observers: dict[str, list[NotifyHandler]] = {}

    def attach(self, notify_type: str, handler: NotifyHandler) -> None:
        """Add observer."""
        if notify_type not in self._observers:
            self._observers[notify_type] = []

        if handler not in self._observers[notify_type]:
            self._observers[notify_type].append(handler)

    def detach(self, notify_type: str, handler: NotifyHandler) -> None:
        """Remove observer."""
        if notify_type not in self._observers:
            return
        if handler in self._observers[notify_type]:
            self._observers[notify_type].remove(handler)

    def notify(self, notify_type: str, notify_parameters: dict[str, Any]) -> None:
        """Call the observers."""
        # Dup obervers list; add/remove could be called when invoking the observers
        observers = list(self._observers.get(notify_type, []))
        for observer in observers:
            try:
                observer(**notify_parameters)
            except Exception as exc:  # pylint: disable=broad-except
                LOG.exception(exc)



"""Definition of an ElkM1 Keypad."""

import datetime as dt

from .connection import Connection
from .const import FunctionKeys, KeypadKeys, Max, TextDescriptions
from .elements import Element, Elements
from .message import ka_encode, kf_encode
from .notify import Notifier


class Keypad(Element):
    """Class representing an Keypad"""

    def __init__(self, index: int, connection: Connection, notifier: Notifier) -> None:
        super().__init__(index, connection, notifier)
        self.area = -1
        self.temperature = -40
        self.last_user_time = dt.datetime.now(dt.UTC)
        self.last_user = -1
        self.code = ""
        self.last_keypress: tuple[str, int] | None = None
        self.last_function_key = FunctionKeys.FORCE_KF_SYNC

    def press_function_key(self, functionkey: FunctionKeys) -> None:
        """(Helper) Send a function key (1, ... 6, *, C)"""
        self._connection.send(kf_encode(self.index, functionkey))


class Keypads(Elements[Keypad]):
    """Handling for multiple areas"""

    def __init__(self, connection: Connection, notifier: Notifier) -> None:
        super().__init__(connection, notifier, Keypad, Max.KEYPADS.value)
        notifier.attach("IC", self._ic_handler)
        notifier.attach("KA", self._ka_handler)
        notifier.attach("KC", self._kc_handler)
        notifier.attach("KF", self._kf_handler)
        notifier.attach("LW", self._lw_handler)
        notifier.attach("ST", self._st_handler)

    def sync(self) -> None:
        """Retrieve areas from ElkM1"""
        self._connection.send(ka_encode())
        self.get_descriptions(TextDescriptions.KEYPAD.value)
        # Send KF for one of our keypads which reports them all
        self._connection.send(kf_encode(0))

    def _ic_handler(self, code: int, user: int, keypad: int) -> None:
        keypad_ = self.elements[keypad]

        # By setting a time this will force the IC change to always be reported
        keypad_.setattr("last_user_time", dt.datetime.now(dt.UTC), False)

        # If user is negative then invalid code entered
        keypad_.setattr("code", code if user < 0 else "****", False)
        keypad_.setattr("last_user", user, True)

    def _ka_handler(self, keypad_areas: list[int]) -> None:
        for keypad in self.elements:
            if keypad_areas[keypad.index] >= 0:
                keypad.setattr("area", keypad_areas[keypad.index], True)

    def _kc_handler(self, keypad: int, key: int) -> None:
        """
        Handle a keypad change message. At present this function is only handling
        keypresses and does not process function key light changes and beep changes.
        """

        # Ignore NO_KEY as it is not a key press
        if key == KeypadKeys.NO_KEY.value:
            return

        # Force a change notification
        self.elements[keypad].last_keypress = None
        try:
            name = KeypadKeys(key).name
        except ValueError:
            name = ""
        self.elements[keypad].setattr("last_keypress", (name, key), True)

    def _kf_handler(self, keypad: int, key: str, chime_mode: list[int]) -> None:
        # Force a change notification
        self.elements[keypad].last_function_key = FunctionKeys.FORCE_KF_SYNC
        try:
            name = FunctionKeys(key).name
        except ValueError:
            name = ""
        self.elements[keypad].setattr("last_function_key", (name, key), True)

    def _lw_handler(self, keypad_temps: list[int], zone_temps: list[int]) -> None:
        for keypad in self.elements:
            if keypad_temps[keypad.index] > -40:
                keypad.setattr("temperature", keypad_temps[keypad.index], True)

    def _st_handler(self, group: int, device: int, temperature: int) -> None:
        if group == 1:
            self.elements[device].setattr("temperature", temperature, True)



import pytest

from elkm1_lib.util import parse_url, pretty_const


def test_parse_url_valid_non_ssl_socket():
    (scheme, host, port, ssl) = parse_url("elk://some.host:1234")
    assert scheme == "elk"
    assert host == "some.host"
    assert port == 1234
    assert ssl is None


def test_parse_url_valid_ssl_socket():
    (scheme, host, port, ssl) = parse_url("elks://another.host:5678")
    assert scheme == "elks"
    assert host == "another.host"
    assert port == 5678
    assert ssl is not None


def test_parse_url_default_non_ssl_port():
    (scheme, host, port, ssl) = parse_url("elk://192.168.0.123")
    assert scheme == "elk"
    assert host == "192.168.0.123"
    assert port == 2101
    assert ssl is None


def test_parse_url_default_ssl_port():
    (scheme, host, port, ssl) = parse_url("elks://ssl.host")
    assert scheme == "elks"
    assert host == "ssl.host"
    assert port == 2601
    assert ssl is not None


def test_parse_url_serial_non_implemented():
    (scheme, host, port, ssl) = parse_url("serial:///dev/tty:4800")
    assert scheme == "serial"
    assert host == "/dev/tty"
    assert port == 4800
    assert ssl is None


def test_parse_url_unknown_scheme():
    with pytest.raises(ValueError):
        parse_url("bad_scheme://rest")


def test_pretty_const_with_single_word():
    rsp = pretty_const("TESTING")
    assert rsp == "Testing"


def test_pretty_const_with_two_words():
    rsp = pretty_const("FIRE_ALARM")
    assert rsp == "Fire alarm"



"""Definition of an ElkM1 Area"""

from __future__ import annotations

from typing import Any

from .connection import Connection
from .const import (
    AlarmState,
    ArmedStatus,
    ArmLevel,
    ArmUpState,
    ChimeMode,
    Max,
    TextDescriptions,
)
from .elements import Element, Elements
from .message import al_encode, as_encode, az_encode, dm_encode, zb_encode
from .notify import Notifier


class Area(Element):
    """Class representing an Area"""

    def __init__(self, index: int, connection: Connection, notifier: Notifier) -> None:
        super().__init__(index, connection, notifier)
        self.armed_status: ArmedStatus | None = None
        self.arm_up_state: ArmUpState | None = None
        self.alarm_state: AlarmState | None = None
        self.alarm_memory = False
        self.is_exit = False
        self.timer1 = 0
        self.timer2 = 0
        self.last_log: str | None = None
        self.chime_mode = None

    def is_armed(self) -> bool:
        """Return if the area is armed."""
        if self.armed_status is None:
            return False
        return self.armed_status != ArmedStatus.DISARMED

    def in_alarm_state(self) -> bool:
        """Return if area is in alarm state."""
        return self.alarm_state not in {
            None,
            AlarmState.NO_ALARM_ACTIVE,
            AlarmState.ENTRANCE_DELAY_ACTIVE,
            AlarmState.ALARM_ABORT_DELAY_ACTIVE,
        }

    def arm(self, level: ArmLevel, code: int) -> None:
        """(Helper) Arm system at specified level (away, vacation, etc)"""
        if self.is_armed() and level != ArmLevel.DISARM:
            return
        self._connection.send(al_encode(level, self._index, code))

    def disarm(self, code: int) -> None:
        """(Helper) Disarm system."""
        self.arm(ArmLevel.DISARM, code)

    def display_message(
        self, clear: int, beep: bool, timeout: int, line1: str, line2: str
    ) -> None:
        """(Helper) Display a message on all of the keypads in this area."""
        self._connection.send(
            dm_encode(self._index, clear, beep, timeout, line1, line2)
        )

    def bypass(self, code: int) -> None:
        """(Helper) Bypass area."""
        self._connection.send(zb_encode(999, self._index, code))

    def clear_bypass(self, code: int) -> None:
        """(Helper) Clear bypass area."""
        self._connection.send(zb_encode(-1, self._index, code))


class Areas(Elements[Area]):
    """Handling for multiple areas"""

    def __init__(self, connection: Connection, notifier: Notifier) -> None:
        super().__init__(connection, notifier, Area, Max.AREAS.value)
        notifier.attach("AM", self._am_handler)
        notifier.attach("AS", self._as_handler)
        notifier.attach("EE", self._ee_handler)
        notifier.attach("KF", self._kf_handler)
        notifier.attach("LD", self._ld_handler)

    def sync(self) -> None:
        """Retrieve areas from ElkM1"""
        self._connection.send(as_encode())
        self.get_descriptions(TextDescriptions.AREA.value)

    def _am_handler(self, alarm_memory: list[bool]) -> None:
        for area in self.elements:
            area.setattr("alarm_memory", alarm_memory[area.index], True)

    def _as_handler(
        self,
        armed_statuses: list[ArmedStatus],
        arm_up_states: list[ArmUpState],
        alarm_states: list[AlarmState],
    ) -> None:
        update_alarm_triggers = False
        for area in self.elements:
            area.setattr("armed_status", armed_statuses[area.index], False)
            area.setattr("arm_up_state", arm_up_states[area.index], False)
            if (
                area.alarm_state != alarm_states[area.index]
                or alarm_states[area.index] != AlarmState.NO_ALARM_ACTIVE
            ):
                update_alarm_triggers = True
            area.setattr("alarm_state", alarm_states[area.index], True)

        if update_alarm_triggers:
            self._connection.send(az_encode())

    def _ee_handler(
        self,
        area: int,
        is_exit: bool,
        timer1: int,
        timer2: int,
        armed_status: ArmedStatus,
    ) -> None:
        area_element = self.elements[area]
        area_element.setattr("armed_status", armed_status, False)
        area_element.setattr("timer1", timer1, False)
        area_element.setattr("timer2", timer2, False)
        area_element.setattr("is_exit", is_exit, True)

    # ElkM1 global setting G35 must be set for LD messages to be sent
    def _ld_handler(self, area: int, log: dict[str, Any]) -> None:
        if log["event"] in [1173, 1174]:
            # arm/disarm log (YAGNI - decode number for more log types when needed)
            log["user_number"] = log["number"]
        self.elements[area].setattr("last_log", log, True)

    def _kf_handler(self, keypad: int, key: str, chime_mode: list[int]) -> None:
        for area, mode in enumerate(chime_mode):
            try:
                name = ChimeMode(mode).name
            except ValueError:
                name = ""
            self.elements[area].setattr("chime_mode", (name, mode), True)



"""Definition of an ElkM1 Output"""

from .connection import Connection
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import cf_encode, cn_encode, cs_encode, ct_encode
from .notify import Notifier


class Output(Element):
    """Class representing an Output"""

    def __init__(self, index: int, connection: Connection, notifier: Notifier) -> None:
        super().__init__(index, connection, notifier)
        self.output_on = False

    def turn_off(self) -> None:
        """(Helper) Turn of an output"""
        self._connection.send(cf_encode(self._index))

    def turn_on(self, time: int) -> None:
        """(Helper) Turn on an output"""
        self._connection.send(cn_encode(self._index, time))

    def toggle(self) -> None:
        """(Helper) Toggle an output"""
        self._connection.send(ct_encode(self._index))


class Outputs(Elements[Output]):
    """Handling for multiple areas"""

    def __init__(self, connection: Connection, notifier: Notifier) -> None:
        super().__init__(connection, notifier, Output, Max.OUTPUTS.value)
        notifier.attach("CC", self._cc_handler)
        notifier.attach("CS", self._cs_handler)

    def sync(self) -> None:
        """Retrieve areas from ElkM1"""
        self._connection.send(cs_encode())
        self.get_descriptions(TextDescriptions.OUTPUT.value)

    def _cc_handler(self, output: int, output_status: bool) -> None:
        self.elements[output].setattr("output_on", output_status, True)

    def _cs_handler(self, output_status: list[bool]) -> None:
        for output in self.elements:
            output.setattr("output_on", output_status[output.index], True)



from unittest.mock import Mock

import pytest

from elkm1_lib.const import SettingFormat
from elkm1_lib.message import MessageEncode
from elkm1_lib.settings import Setting, Settings

from .util import rx_msg


@pytest.fixture
def settings(notifier):
    return Settings(Mock(), notifier)


def test_settings_number(settings, notifier):
    rx_msg("CR", "01001230", notifier)
    setting = settings[0]
    assert setting.value_format == SettingFormat.NUMBER
    assert setting.value == 123


def test_settings_time_of_day(settings, notifier):
    rx_msg("CR", "01054162", notifier)
    setting = settings[0]
    assert setting.value_format == SettingFormat.TIME_OF_DAY
    assert setting.value == (21, 40)


def test_setting_set_types_are_correct():
    setting = Setting(0, Mock(), Mock())

    setting.value_format = SettingFormat.NUMBER
    with pytest.raises(ValueError):
        setting.set((1, 2))

    setting.value_format = SettingFormat.TIME_OF_DAY
    with pytest.raises(ValueError):
        setting.set(42)


def test_setting_write_new_value_called():
    mock = Mock()
    setting = Setting(0, mock, Mock())

    setting.value_format = SettingFormat.NUMBER
    setting.set(42)
    mock.send.assert_called_with(
        MessageEncode(message="0Dcw010004200", response_command=None)
    )

    setting.value_format = SettingFormat.TIME_OF_DAY
    setting.set((21, 40))
    mock.send.assert_called_with(
        MessageEncode(message="0Dcw010541600", response_command=None)
    )



#!/usr/bin/env python

import logging
import os

from elkm1_lib import Elk
from elkm1_lib.message import ua_encode

LOG = logging.getLogger(__name__)


class MyApp:
    def __init__(self, url, user_code):
        # Setting element list to empty list means no elements are synced on startup
        # i.e.: saves startup sync time if you don't need it!
        self.elk = Elk({"url": url, "element_list": []})

        # Add some handlers here which get called when their respective event occurs.
        self.elk.add_handler("disconnected", self.disconnected)
        self.elk.add_handler("connected", self.connected)
        self.elk.add_handler("sync_complete", self.sync_complete)
        self.elk.connect()

        self.user_code = user_code

        # Start the asyncio event loop.
        self.elk.run()

    def disconnected(self):
        """The ElkM1 panel has disconnect."""
        print("Disconnected!!!")
        exit(0)  # Just exit since this is a demo, might want to try and reconnect.

    def connected(self):
        """The ElkM1 is connected - we can communicate with the panel now!"""
        print("Connected!!!")

    def sync_complete(self):
        """
        On startup of the lib, we synchronize the status of all the elements on
        the panel to the lib. Things such as zone status, temperature, etc. This
        handler is called when the sychronization is complete. It's a good spot
        for your app to sending its own messages to the ElkM1.
        """
        print("Sync complete!!!")

        # Add a method that is called when a response to a UA message is received.
        self.elk.add_handler("UA", self.ua_response)

        # Send a UA message
        self.elk.send(ua_encode(self.user_code))

    def ua_response(self, **kwargs):
        """Got a response to our UA message!"""
        print(kwargs)

        # Got the UA response! Add your own app code here.
        # We just disconnect since we got the one response we want.
        self.elk.disconnect()


def main():
    logging.basicConfig(level=logging.DEBUG, format="%(message)s")

    # Uses an environment variable for simplicity; could be command line argument
    url = os.environ.get("ELKM1_URL")
    if not url:
        print("Specify url to connect to in ELKM1_URL environment variable")
        exit(0)
    _ = MyApp(url, 8813)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        exit(0)



"""Master class that combines all ElkM1 pieces together."""

from __future__ import annotations

import asyncio
import logging
from typing import Any

from .areas import Areas
from .connection import Connection
from .counters import Counters
from .keypads import Keypads
from .lights import Lights
from .message import MessageEncode, MsgHandler, ua_encode
from .notify import Notifier
from .outputs import Outputs
from .panel import Panel
from .settings import Settings
from .tasks import Tasks
from .thermostats import Thermostats
from .users import Users
from .util import url_scheme_is_secure
from .zones import Zones

LOG = logging.getLogger(__name__)


class Elk:
    """Represents all the components on an Elk panel."""

    def __init__(
        self, config: dict[str, Any], loop: asyncio.AbstractEventLoop | None = None
    ) -> None:
        """Initialize a new Elk instance."""
        self._config = config
        if not loop:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
        self._loop = loop

        self._notifier = Notifier()
        self._connection = Connection(config["url"], self._notifier)
        self._logged_in = False

        # Setup for all the types of elements tracked
        if "element_list" in config:
            self.element_list = config["element_list"]
        else:
            self.element_list = [
                "panel",
                "zones",
                "lights",
                "areas",
                "tasks",
                "keypads",
                "outputs",
                "thermostats",
                "counters",
                "settings",
                "users",
            ]

        self._notifier.attach("connected", self._connected)
        self._notifier.attach("disconnected", self._disconnected)
        self._notifier.attach("login", self._login_status)
        self._notifier.attach("IE", self._call_sync_handlers)
        self._notifier.attach("VN", self._got_first_message)

        self.areas = Areas(self._connection, self._notifier)
        self.counters = Counters(self._connection, self._notifier)
        self.keypads = Keypads(self._connection, self._notifier)
        self.lights = Lights(self._connection, self._notifier)
        self.outputs = Outputs(self._connection, self._notifier)
        self.panel = Panel(self._connection, self._notifier)
        self.settings = Settings(self._connection, self._notifier)
        self.tasks = Tasks(self._connection, self._notifier)
        self.thermostats = Thermostats(self._connection, self._notifier)
        self.users = Users(self._connection, self._notifier)
        self.zones = Zones(self._connection, self._notifier)

    def _login_status(self, succeeded: bool) -> None:
        self._logged_in = succeeded
        if not succeeded:
            self._connection.disconnect()
            LOG.error("Invalid username or password.")

    def _got_first_message(self, **kwargs: dict[str, Any]) -> None:
        if not self._logged_in:
            self._notifier.notify("login", {"succeeded": True})

    def _connected(self) -> None:
        if url_scheme_is_secure(self._config["url"]):
            self._connection.send_raw(self._config["userid"])
            self._connection.send_raw(self._config["password"])
        self._call_sync_handlers()

    def _disconnected(self) -> None:
        self._logged_in = False

    def _sync_complete(self, **_: dict[str, Any]) -> None:
        self._notifier.notify("sync_complete", {})

        # Remove so that other apps can send UA and not trigger sync_complete
        self._notifier.detach("UA", self._sync_complete)

    def _call_sync_handlers(self) -> None:
        """Invoke the synchronization handlers."""

        LOG.debug("Synchronizing panel...")
        self.add_handler("UA", self._sync_complete)
        for element in self.element_list:
            getattr(self, element).sync()
        self.send(ua_encode(0))  # Used to mark end of sync

    @property
    def connection(self) -> Connection:
        """Return the connection instance."""
        return self._connection

    def run(self) -> None:
        """Enter the asyncio loop."""
        self._loop.run_forever()

    def add_handler(self, msg_type: str, handler: MsgHandler) -> None:
        """Helper to connection add_handler."""
        self._notifier.attach(msg_type, handler)

    def remove_handler(self, msg_type: str, handler: MsgHandler) -> None:
        """Helper to connection remove_handler."""
        self._notifier.detach(msg_type, handler)

    def connect(self) -> None:
        """Helper to connection connect."""
        asyncio.ensure_future(self._connection.connect())

    def disconnect(self) -> None:
        """Helper to connection disconnect."""
        self._connection.disconnect()

    def is_connected(self) -> bool:
        """Helper to connection is_connected."""
        return self._connection.is_connected()

    def is_paused(self) -> bool:
        """Helper to connection is_paused."""
        return self._connection.is_paused()

    def send(self, msg: MessageEncode) -> None:
        """Helper to connection send."""
        self._connection.send(msg)



from unittest.mock import Mock

import pytest

from elkm1_lib.areas import Area, Areas
from elkm1_lib.const import AlarmState, ArmedStatus, ArmUpState
from elkm1_lib.message import MessageEncode

from .util import rx_msg


@pytest.fixture
def areas(notifier):
    return Areas(Mock(), notifier)


def test_entry_exit_general(areas, notifier):
    rx_msg("EE", "100601201", notifier)
    area = areas[0]
    assert area.is_exit is True
    assert area.timer1 == 60
    assert area.timer2 == 120
    assert area.armed_status == ArmedStatus.ARMED_AWAY


def test_entry_exit_armed_status(areas, notifier):
    area = areas[0]
    assert area.is_armed() is False
    rx_msg("EE", "110300601", notifier)
    assert area.is_armed() is True


def test_alarm_memory(areas, notifier):
    rx_msg("AM", "10101010", notifier, zeros="")
    assert areas[0].alarm_memory is True
    assert areas[1].alarm_memory is False


def test_armed_status(areas, notifier):
    rx_msg("AS", "100000004000000030000000", notifier)
    area = areas[0]
    assert area.armed_status == ArmedStatus.ARMED_AWAY
    assert area.arm_up_state == ArmUpState.FULLY_ARMED
    assert area.alarm_state == AlarmState.FIRE_ALARM


def test_armed_status_updates_alarm_triggers(notifier):
    mock_connection = Mock()
    Areas(mock_connection, notifier)
    rx_msg("AS", "100000004000000030000000", notifier)
    mock_connection.send.assert_called_with(
        MessageEncode(message="06az00", response_command="AZ")
    )


def test_is_in_alarm_state():
    area = Area(0, Mock(), Mock())
    assert area.in_alarm_state() is False

    area.alarm_state = AlarmState.POLICE_ALARM
    assert area.in_alarm_state() is True



"""
Created on Aug 2, 2015
From: https://github.com/izderadicka/xmpp-tester/blob/master/commander.py
@author: ivan
"""

import inspect
import re
import threading
import traceback
from collections import deque
from importlib import import_module

import attr
import urwid

import elkm1_lib.message


@attr.s
class Command:
    function = attr.ib()
    help = attr.ib()
    docs = attr.ib()


def parse_subcommand(line):
    tokens = line.split()
    if len(tokens) == 0:
        return (None, None)
    cmd = tokens[0].lower()
    args = tokens[1:]

    converted = []
    for arg in args:
        try:
            i = int(arg)
            converted.append(i)
        except ValueError:
            converted.append(arg)
    return (cmd, converted)


def parse_range(rng, max):
    # e.g. rng = "<3, 45, 46, 48-51, 77"
    if rng.strip() == "*":
        return range(0, max)
    ids = []
    for x in map(str.strip, rng.split(",")):
        if x.isdigit():
            ids.append(int(x))
        elif x[0] == "<":
            ids.extend(range(0, int(x[1:])))
        elif "-" in x:
            xr = [s.strip() for s in x.split("-")]
            ids.extend(range(int(xr[0]), int(xr[1]) + 1))
        else:
            raise ValueError(f'Unknown range type: "{x}"')
    return ids


def parse_element_command(cmd, line, max):
    match = re.match("([\\d,\\- <*]+)(\\w*.*)", line)
    if match:
        ids = parse_range(match.groups()[0], max)
        subcommand = parse_subcommand(match.groups()[1])
        return (ids, subcommand[0], subcommand[1])
    return None


def find_class(module, class_name):
    for class_ in inspect.getmembers(module, inspect.isclass):
        if class_[0] == class_name:
            return class_
    return None


def get_helpers(element, clas):
    helpers = {}
    class_ = find_class(import_module("elkm1_lib." + element), clas)
    if class_:
        for function_name, fn in inspect.getmembers(class_[1], inspect.isfunction):
            if fn.__doc__ and fn.__doc__.startswith("(Helper)"):
                params = [p for p in inspect.signature(fn).parameters]
                if params[0] == "self":
                    del params[0]
                params = " ".join(["<" + p + ">" for p in params])
                helpers[function_name] = (
                    fn,
                    f"{function_name} {params}",
                    fn.__doc__[8:],
                )
    return helpers


class Commands:
    def __init__(self, elk):
        self.elk = elk

        self._quit_cmd = ["q", "quit", "exit"]
        self._help_cmd = ["?", "help", "h"]

        self.encode_cmds = {}
        for fn_name, fn in inspect.getmembers(elkm1_lib.message, inspect.isfunction):
            if not fn_name.endswith("_encode"):
                continue
            cmd = fn_name[0:2]
            params = [p for p in inspect.signature(fn).parameters]
            params = " ".join(["<" + p + ">" for p in params])
            self.encode_cmds[cmd] = Command(fn, f"{cmd} {params}", fn.__doc__)

        self.element_cmds = {}
        self.subcommands = {}
        for element in elk.element_list:
            if element == "panel":
                fn = self.panel_print
                cmd = element
            else:
                fn = self.print_elements
                cmd = element[:-1]
            self.element_cmds[cmd] = (
                fn,
                f"{cmd} <range list> [subcommand]",
                f"Displays internal state of {element}",
                get_helpers(element, cmd.capitalize()),
            )

    def __call__(self, line):
        tokens = line.split()
        cmd = tokens[0].lower()
        args = tokens[1:]

        if cmd in self._quit_cmd:
            return Commander.Exit

        print(f"#blue#{line}")

        if cmd in self._help_cmd:
            return self.help(cmd, args)

        if cmd in self.encode_cmds:
            return self.encoder(cmd, args)

        if cmd == "recv":
            return self.elk._got_data(" ".join(args))

        if cmd in self.element_cmds:
            return self.element_cmds[cmd][0](cmd, args)

        return f"#error#Unknown command: {cmd}"

    def help(self, cmd, args):
        if len(args) == 0:
            res = '#green#Type "[?|h|help] <command>" to get more help\n'
            res += 'Type "[q|quit|exit]" to quit program\n'
            res += "Element display commands:\n  {}\n\n".format(
                " ".join(list(self.element_cmds.keys()))
            )
            cl = [fnname for fnname in self.encode_cmds]
            res += "Send message commands:\n  {}\n".format(" ".join(sorted(cl)))
        else:
            help_for = args[0]
            if help_for in self.encode_cmds:
                command = self.encode_cmds[help_for]
                res = f"#green#{command.help}\n{command.docs}"

            elif help_for in self.element_cmds:
                res = (
                    f"#green#{self.element_cmds[help_for][1]}\n"
                    f"{self.element_cmds[help_for][2]}"
                )
                for _k, v in self.element_cmds[help_for][3].items():
                    res += f"\nSubcommand: {v[1]}\n{v[2]}"
            else:
                res = f"#error#Unknown command: {help_for}"
        return res

    def print_elements(self, cmd, args):
        element_list = getattr(self.elk, cmd + "s", None)
        if element_list is None:
            print("Command not supported.")
            return

        args = parse_element_command(cmd, " ".join(args), element_list.max_elements)
        if args is None:
            print("Cannot parse command.")
            return

        if args[1]:
            if args[1] in self.element_cmds[cmd][3]:
                fn = self.element_cmds[cmd][3][args[1]][0]
            else:
                raise NotImplementedError
            for i in args[0]:
                print(fn)
                fn(element_list[i], *args[2])
        else:
            for i in args[0]:
                print(element_list[i])

    def panel_print(self, cmd, args):
        print(self.elk.panel)

    def encoder(self, cmd, args):
        converted = []
        for arg in args:
            try:
                i = int(arg)
                converted.append(i)
            except ValueError:
                converted.append(arg)
        self.elk.send(self.encode_cmds[cmd].function(*converted))


class FocusMixin:
    def mouse_event(self, size, event, button, x, y, focus):
        if focus and hasattr(self, "_got_focus") and self._got_focus:
            self._got_focus()
        return super().mouse_event(size, event, button, x, y, focus)


class ListView(FocusMixin, urwid.ListBox):
    def __init__(self, model, got_focus, max_size=None):
        urwid.ListBox.__init__(self, model)
        self._got_focus = got_focus
        self.max_size = max_size
        self._lock = threading.Lock()

    def mouse_event(self, size, event, button, x, y, focus):
        direction = "up" if button == 4 else "down"
        return super().keypress(size, direction)

    def add(self, line):
        with self._lock:
            was_on_end = self.get_focus()[1] == len(self.body) - 1
            if self.max_size and len(self.body) > self.max_size:
                del self.body[0]
            self.body.append(urwid.Text(line))
            last = len(self.body) - 1
            if was_on_end:
                self.set_focus(last, "above")


class Input(FocusMixin, urwid.Edit):
    signals = ["line_entered"]

    def __init__(self, got_focus=None):
        urwid.Edit.__init__(self)
        self.history = deque(maxlen=1000)
        self._history_index = -1
        self._got_focus = got_focus

    def keypress(self, size, key):
        if key == "enter":
            line = self.edit_text.strip()
            if line:
                urwid.emit_signal(self, "line_entered", line)
                self.history.append(line)
            self._history_index = len(self.history)
            self.edit_text = ""
        if key == "up":
            self._history_index -= 1
            if self._history_index < 0:
                self._history_index = 0
            else:
                self.edit_text = self.history[self._history_index]
        if key == "down":
            self._history_index += 1
            if self._history_index >= len(self.history):
                self._history_index = len(self.history)
                self.edit_text = ""
            else:
                self.edit_text = self.history[self._history_index]
        else:
            urwid.Edit.keypress(self, size, key)


class Commander(urwid.Frame):
    """Simple terminal UI with command input on bottom line and display
    frame above similar to chat client etc. Initialize with your
    Command instance to execute commands and the start main loop
    Commander.loop(). You can also asynchronously output messages
    with Commander.output('message')"""

    class Exit:
        pass

    PALLETE = [
        ("reversed", urwid.BLACK, urwid.LIGHT_GRAY),
        ("normal", urwid.LIGHT_GRAY, urwid.BLACK),
        ("error", urwid.LIGHT_RED, urwid.BLACK),
        ("green", urwid.DARK_GREEN, urwid.BLACK),
        ("blue", urwid.LIGHT_BLUE, urwid.BLACK),
        ("magenta", urwid.DARK_MAGENTA, urwid.BLACK),
    ]

    def __init__(
        self,
        title,
        command_caption="Command:  (Tab to switch focus to upper frame)",
        cmd_cb=None,
        max_size=1000,
    ):
        self.header = urwid.Text(title)
        self.model = urwid.SimpleListWalker([])
        self.body = ListView(
            self.model, lambda: self._update_focus(False), max_size=max_size
        )
        self.input = Input(lambda: self._update_focus(True))
        foot = urwid.Pile(
            [
                urwid.AttrMap(urwid.Text(command_caption), "reversed"),
                urwid.AttrMap(self.input, "normal"),
            ]
        )
        urwid.Frame.__init__(
            self,
            urwid.AttrWrap(self.body, "normal"),
            urwid.AttrWrap(self.header, "reversed"),
            foot,
        )
        self.set_focus_path(["footer", 1])
        self._focus = True
        urwid.connect_signal(self.input, "line_entered", self.on_line_entered)
        self._cmd = cmd_cb
        self._output_styles = [s[0] for s in self.PALLETE]
        self.eloop = None

    def loop(self, handle_mouse=False):
        # self.eloop=urwid.MainLoop(self, self.PALLETE, handle_mouse=handle_mouse)
        # self._eloop_thread=threading.current_thread()
        # self.eloop.run()
        import asyncio

        evl = urwid.AsyncioEventLoop(loop=asyncio.get_event_loop())
        self.eloop = urwid.MainLoop(
            self, self.PALLETE, event_loop=evl, handle_mouse=True
        )
        self._eloop_thread = threading.current_thread()
        self.eloop.run()

    def on_line_entered(self, line):
        if self._cmd:
            try:
                res = self._cmd(line)
            except Exception as e:
                traceback.print_exc()
                self.output(f"Error: {e}", "error")
                return
            if res == Commander.Exit:
                raise urwid.ExitMainLoop()
            elif res:
                self.output(str(res))
        else:
            if line in ("q", "quit", "exit"):
                raise urwid.ExitMainLoop()
            else:
                self.output(line)

    def output(self, line, style=None):
        match = re.search(r"^#(\w+)#(.*)", line, re.DOTALL)
        if match and match.group(1) in self._output_styles:
            line = (match.group(1), match.group(2))
        elif style and style in self._output_styles:
            line = (style, line)
        self.body.add(line)

        # since output could be called asynchronously form other threads
        # we need to refresh screen in these cases
        if self.eloop and self._eloop_thread != threading.current_thread():
            self.eloop.draw_screen()

    def _update_focus(self, focus):
        self._focus = focus

    def switch_focus(self):
        if self._focus:
            self.set_focus("body")
            self._focus = False
        else:
            self.set_focus_path(["footer", 1])
            self._focus = True

    def keypress(self, size, key):
        if key == "tab":
            self.switch_focus()
        return urwid.Frame.keypress(self, size, key)
